﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectから取得した骨格情報を基にセンサ値を残像表示する』
//　DepthImage、ColorImageの人物切り取り表示は行わない
//【名称】Bitmap_GetSkeletonData　　【Version】1.02
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年3月5日　【コメント追記】2012年7月2日
//【注意事項】
//　このプログラムをKinect for Windows SDK Version1.5で動かすには、参照設定の
// Microsoft.Kinectを一旦削除し、プロジェクト(P)-参照の追加(F)で、.NETタブ内
// にあるMicrosoft.Kinectを選択・追加してください。 
//*************************************************************************************************

using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;         //【BitmapData】
using System.Diagnostics;             //【Debug】
using System.Runtime.InteropServices; //【Marshal.Copy】【Win32.API】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Kinect_ARsensing
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private ColorImageFrame KS_ColorFrame;  //【KINECTのColorImageFrame】
        private byte[] KS_ColorImageDataARGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8で受けてA8R8G8B8に変換
        private int KS_ColorFrameWidth;         //【KINECTのColorFrameの横方向画素数】
        private int KS_ColorFrameHeight;        //【KINECTのColorFrameの縦方向画素数】
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private DepthImageFrame KS_DepthFrame;  //【KINECTのDepthImageFrame】
        private short[] KS_DepthData;           //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataXRGB;   //【KINECTのDepthImageData】32bit/画素,X8R8G8B8表示形式
        private int KS_DepthFrameWidth;         //【KINECTのDepthFrameの横方向画素数】
        private int KS_DepthFrameHeight;        //【KINECTのDepthFrameの縦方向画素数】
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】
        private Rectangle KS_rect2;             //【DepthImage表示枠】
        private int KS_PlayerIndexBitmask;      //【DepthデータのPlayer情報】 0000 0000 0000 0111B = 7
        private int KS_PlayerIndexBitmaskWidth; //【DepthデータのPlayer情報】 3 bits
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private SkeletonFrame KS_SkeletonFrame; //【KINECTのSkeletonFrame】
        private Skeleton[] KS_SkeletonData;     //【KINECTのSkeleton】Skeleton 最大6名、20Joints/Player
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private bool KS_flagPlayer = false;     //【Player検出フラグ】
        private long KS_TimeStamp;              // ImageFrameのタイムスタンプ

        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar ColorBar = new NTL_ColorBar(); //【デプス値をカラーに変換する部品】
        private NTL_SkeletonData NTL_SkeletonData;
        private const int nFIFO = 60; //【ジェスチャー推定処理用FIFOの段数】2秒(60フレーム)分
        private SkeletonPoint Ps;   //working用
        private ColorImagePoint Pc; //working用
        private NTL_EstimationCircle estR = new NTL_EstimationCircle();
        private NTL_EstimationCircle estL = new NTL_EstimationCircle();
        private NTL_EstimationCircle.NTL_FigureType FigR, FigL; //検出状態
        private int DetectCountR = 0; //一瞬のLine判定の後、５秒間は無視。
        private int DetectCountL = 0; //一瞬のLine判定の後、５秒間は無視。
        private Boolean NTL_flagGesture = true;
        private Boolean NTL_flagSkeleton = true;

        //【USB ADC基板のコントロール】
        private String COMPORT; //【A/Dコンバータ通信用ComPort番号】←ここはデバイス検出状況に応じて変更してください。
        System.IO.Ports.SerialPort SerialPort_for_ADC = new System.IO.Ports.SerialPort(); //【ADC用シリアルポート】
        private string RxMsg; //【受信文字列】
        private Boolean NTL_WaveDisplay = true;
        private int ADcount = 0;
        private float SensorValue;
        private string sSensorValue = "";
        private int MeasureValue = 0;
        private float[] NTL_WaveFIFO = new float[640];
        private int NTL_Wave_wPtr = 0;
        private int NTL_Wave_rPtr = 0;
        private int NTL_Ch = 0;
        private int NTL_pCh = 0;

        //【センサ・プローブの測定値の履歴を管理するためのFIFO】
        private const int NTL_default_nFIFO = 300; //【FIFO段数】Default 30FPSで10秒分に相当
        private int NTL_nFIFO = NTL_default_nFIFO; //【FIFO段数】
        private int NTL_wPtr;                      //【FIFO書込みポインタ】
        private int NTL_rPtr;                      //【FIFO読出しポインタ】
        private bool NTL_FIFOfull = false;         //【FIFOフル】
        private bool NTL_FIFOempty = true;         //【FIFO空】
        private NTL_FIFOData[] NTL_FIFO_Sensor = new NTL_FIFOData[NTL_default_nFIFO];
        private NTL_FIFOData NTL_HandySensor;        //【右手】か【左手】
        private NTL_ColorBar NTL_SensorColorBar = new NTL_ColorBar();

        public struct NTL_FIFOData
        {
            public SkeletonPoint Ps;
            public DepthImagePoint Pd;
            public ColorImagePoint Pc;
            public Color ColorValue;
            public int SensorValue;
        }

        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】ColorStream, DepthStream, SkeletonStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution320x240Fps30);    //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                        //【SkeltonStream許可】
            KS_KinectSensor.Start();                                                        //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(ColorImageReady); //【カラー画像取得準備完了時のcall先】
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(DepthImageReady); //【奥行き情報取得準備完了時のcall先】
            KS_KinectSensor.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KinectAllFramesReady);   //【骨格情報取得準備完了時のcall先】
            KS_ColorImageDataARGB = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength]; //【カラー画像データ用】
            KS_ColorFrameWidth = KS_KinectSensor.ColorStream.FrameWidth;   //【ColorImageの横方向画素数】
            KS_ColorFrameHeight = KS_KinectSensor.ColorStream.FrameHeight; //【ColorImageの縦方向画素数】
            KS_bmp1 = new Bitmap(KS_ColorFrameWidth, KS_ColorFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ColorFrameWidth, KS_ColorFrameHeight); //【ColorImage用枠】

            KS_DepthData = new short[KS_KinectSensor.DepthStream.FramePixelDataLength];              //【深度情報データ用】
            KS_DepthImageDataXRGB = new byte[KS_KinectSensor.DepthStream.FramePixelDataLength * 4];  //【深度情報データ表示用】
            KS_DepthFrameWidth = KS_KinectSensor.DepthStream.FrameWidth;   //【DepthImageの横方向画素数】
            KS_DepthFrameHeight = KS_KinectSensor.DepthStream.FrameHeight; //【DepthImageの縦方向画素数】
            KS_bmp2 = new Bitmap(KS_DepthFrameWidth, KS_DepthFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【DepthImage表示用】
            KS_rect2 = new Rectangle(0, 0, KS_DepthFrameWidth, KS_DepthFrameHeight); //【DepthImage用枠】
            KS_PlayerIndexBitmask = DepthImageFrame.PlayerIndexBitmask;
            KS_PlayerIndexBitmaskWidth = DepthImageFrame.PlayerIndexBitmaskWidth;
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Depth測定モードをDefaultに設定】
            KS_CheckDistance();                                        //【設定結果を表示】Lmax, Lminを取得

            KS_SkeletonData = new Skeleton[KS_KinectSensor.SkeletonStream.FrameSkeletonArrayLength]; //【骨格情報用】

            KS_KinectSensor.ElevationAngle = 0; //【Kinectヘッドの仰角を設定】設定範囲は±27°

            ColorBar.CreateColorPalletForAbsolute(KS_Lmax - KS_Lmin + 10); //【カラーパレットを作成する】
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage; //【カラースケールのBitmapイメージがpictureBox枠内にはめ込まれます】
            pictureBox3.Image = ColorBar.CreateColorScale(256, 10); //【256×10画素サイズのカラーバーを作成して表示する】

            checkBox2.CheckedChanged += new EventHandler(checkBox2_CheckedChanged);
            checkBox3.CheckedChanged += new EventHandler(checkBox3_CheckedChanged);
            checkBox4.CheckedChanged += new EventHandler(checkBox4_CheckedChanged);
            pictureBox4.SizeMode = PictureBoxSizeMode.StretchImage; //【デプスイメージがpictureBox枠内にはめ込まれます】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】
            estR.Initialize(nFIFO); //【右手の軌跡から円を推定する準備を行う】
            estL.Initialize(nFIFO); //【左手の軌跡から円を推定する準備を行う】
            NTL_SkeletonData = new NTL_SkeletonData();

            for (int i = 0; i < NTL_WaveFIFO.Length; i++) { NTL_WaveFIFO[i] = 0; }
            string[] pn = System.IO.Ports.SerialPort.GetPortNames();
            if (pn.Length == 0)
            {
                MessageBox.Show(
                   "接続可能なシリアル通信ポートが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            COMPORT = pn[0]; //【一つしかシリアルポートが接続されていない前提で選択】

            InitializeSerialPort(COMPORT);
            ADCInitialize();
            SerialPort_for_ADC.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(SerialPort_for_ADC_DataReceived);
            ADCConvert();  //【次のA/D変換データを要求する】
            radioButton3.CheckedChanged += new EventHandler(radioButton3_CheckedChanged);
            radioButton4.CheckedChanged += new EventHandler(radioButton4_CheckedChanged);
            radioButton5.CheckedChanged += new EventHandler(radioButton5_CheckedChanged);
            radioButton6.CheckedChanged += new EventHandler(radioButton6_CheckedChanged);
            radioButton7.CheckedChanged += new EventHandler(radioButton7_CheckedChanged);
            NTL_FIFO_Initialize(); //【defaltサイズでセンサ用FIFOを作成】
            NTL_SensorColorBar.CreateColorPalletForSensor(256);               //【256色のカラーパレットを作成】
            pictureBox5.SizeMode = PictureBoxSizeMode.StretchImage;           //【カラースケールのBitmapイメージがpictureBox枠内にはめ込まれます】
            pictureBox5.Image = NTL_SensorColorBar.CreateColorScale(256, 10); //【256×10画素サイズのカラーバーを作成して表示する】
        }

        //***************************************************************
        //【RadioButtonが変化した場合に発生するイベント】
        //***************************************************************
        void radioButton3_CheckedChanged(object sender, EventArgs e) { NTL_Ch = 0; } //【超音波センサ】
        void radioButton4_CheckedChanged(object sender, EventArgs e) { NTL_Ch = 1; } //【CdS照度センサ】
        void radioButton5_CheckedChanged(object sender, EventArgs e) { NTL_Ch = 2; } //【NTCサーミスタ】
        void radioButton6_CheckedChanged(object sender, EventArgs e) { NTL_Ch = 3; } //【内蔵温度センサ】
        void radioButton7_CheckedChanged(object sender, EventArgs e) { NTL_Ch = 4; } //【金属検出センサ】


        void checkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox2.Checked) { NTL_flagSkeleton = true; } else { NTL_flagSkeleton = false; }
        }

        void checkBox3_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox3.Checked) { NTL_flagGesture = true; } else { NTL_flagGesture = false; }
        }

        void checkBox4_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox4.Checked) { NTL_WaveDisplay = true; } else { NTL_WaveDisplay = false; }
        }

        //***************************************************************
        //【CheckDistance】Defaultモード(800mm～4000mm)
        //***************************************************************
        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {  //【Defaultモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Default;
            KS_CheckDistance();
        }

        //***************************************************************
        //【距離モード変更】Nearモード(400m～3000mm)
        //***************************************************************
        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {  //【Nearモード】
            KS_KinectSensor.DepthStream.Range = DepthRange.Near;
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【終了要求ボタン】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
            this.Close();              //【Formを閉じる】
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (KS_ColorFrame = e.OpenColorImageFrame())
            {
                if (KS_ColorFrame != null)
                {
                    int i, j, k1, k2, n, xc, yc, r;
                    Graphics g;
                    SolidBrush newbrush;
                    Pen cp;

                    KS_ColorFrame.CopyPixelDataTo(KS_ColorImageDataARGB); //【KINECTのColor画像を配列PixelDataにコピー】
                    for(j=3; j<KS_ColorImageDataARGB.Length; j+=4)
                    {
                        KS_ColorImageDataARGB[j] = 255;
                    }
                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataARGB, 0, KS_data1.Scan0, KS_ColorImageDataARGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;            //【pictureBox1に画像表示】

                    g = Graphics.FromImage(pictureBox1.Image);
                    if (KS_flagPlayer)
                    {

                        if (NTL_flagSkeleton) //【骨格表示処理】
                        {
                            newbrush = new SolidBrush(Color.Red);
                            for (j = 0; j < NTL_SkeletonData.JointPosition.Length; j++)
                            {
                                g.FillEllipse(newbrush, (float)NTL_SkeletonData.JointPosition[j].X - 5f, (float)NTL_SkeletonData.JointPosition[j].Y - 5f, 10f, 10f);
                            }
                            for (j = 0; j < NTL_SkeletonData.Bone.Length; j++)
                            {
                                k1 = NTL_SkeletonData.BoneIndexList[2 * j];
                                k2 = NTL_SkeletonData.BoneIndexList[2 * j + 1];
                                g.DrawLine(new Pen(Color.Yellow, 3.0f),
                                           new PointF((float)NTL_SkeletonData.JointPosition[k1].X, (float)NTL_SkeletonData.JointPosition[k1].Y),
                                           new PointF((float)NTL_SkeletonData.JointPosition[k2].X, (float)NTL_SkeletonData.JointPosition[k2].Y));
                            }
                        }

                        if (NTL_flagGesture) //【ジェスチャー推定処理】
                        {
                            if (estL.Full)
                            {
                                //【左手軌跡の推定円を描画】
                                xc = (int)estL.X; yc = (int)estL.Y; r = (int)estL.Radius;
                                cp = new Pen(Color.DeepSkyBlue, 5.0f);
                                g.DrawEllipse(cp, xc - r, yc - r, 2 * r, 2 * r);
                            }
                            if (estR.Full)
                            {
                                //【右手軌跡の推定円を描画】
                                xc = (int)estR.X; yc = (int)estR.Y; r = (int)estR.Radius;
                                cp = new Pen(Color.DeepPink, 5.0f);
                                g.DrawEllipse(cp, xc - r, yc - r, 2 * r, 2 * r);
                            }
                        }

                        if (NTL_FIFOempty == false)
                        {
                            NTL_FIFOData a;
                            if (NTL_FIFOfull)
                            {
                                // g = Graphics.FromImage(pictureBox1.Image);
                                NTL_rPtr = NTL_wPtr;
                                for (i = 0; i < NTL_nFIFO; i++)
                                {
                                    a = NTL_FIFO_Read_Sensor();
                                    newbrush = new SolidBrush(a.ColorValue);
                                    g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                                }
                            }
                            else
                            {
                                NTL_rPtr = 0;
                                for (i = 0; i < NTL_wPtr; i++)
                                {
                                    a = NTL_FIFO_Read_Sensor();
                                    newbrush = new SolidBrush(a.ColorValue);
                                    g.FillEllipse(newbrush, (float)a.Pc.X - 5f, (float)a.Pc.Y - 5f, 10f, 10f);
                                }
                            }
                        }

                        /*
                        //【右手の残像を描画】
                        newbrush = new SolidBrush(Color.Red);
                        if (estR.Full) { j = estR.Pointer; n = estR.Length; } else { j = 0; n = estR.Pointer; }
                        for (i = 0; i < n; i++)
                        {
                            g.FillEllipse(newbrush, estR.Px[j] - 5f, estR.Py[j] - 5f, 10f, 10f);
                            j++; if (j == n) { j = 0; }
                        }
                        //【左手の残像を描画】
                        newbrush = new SolidBrush(Color.Green);
                        if (estL.Full) { j = estL.Pointer; n = estL.Length; } else { j = 0; n = estL.Pointer; }
                        for (i = 0; i < n; i++)
                        {
                            g.FillEllipse(newbrush, estL.Px[j] - 5f, estL.Py[j] - 5f, 10f, 10f);
                            j++; if (j == n) { j = 0; }
                        }
                        */

                        if (NTL_flagGesture) //【ジェスチャー推定処理】
                        {
                            g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAlias;
                            string s;
                            s = CheckEstimationResult(1, estL.Type);
                            g.DrawString(s, new System.Drawing.Font("New Times Roman", 20.0F), new SolidBrush(Color.Green), 0, 0);
                            s = CheckEstimationResult(0, estR.Type);
                            g.DrawString(s, new System.Drawing.Font("New Times Roman", 20.0F), new SolidBrush(Color.Red), 0, 25);
                        }
                    }

                    //【センサ測定値の波形表示処理】
                    if (NTL_WaveDisplay)
                    {
                        //【ADCの測定値を表示する】
                        this.label4.Text = sSensorValue;
                        NTL_WaveFIFO[NTL_Wave_wPtr] = 300.0f*(1.0f-SensorValue)+100.0f; //【範囲】0.0～1.0⇒100.0f～300.0f
                        NTL_Wave_rPtr = NTL_Wave_wPtr;
                        NTL_Wave_wPtr++; if (NTL_Wave_wPtr == NTL_WaveFIFO.Length) { NTL_Wave_wPtr = 0; }
                        if (pictureBox1.Image != null)
                        {
                            g = Graphics.FromImage(pictureBox1.Image);
                            PointF[] p1 = new PointF[NTL_WaveFIFO.Length];
                            for (i = 0; i < NTL_WaveFIFO.Length; i++)
                            {
                                p1[i] = new PointF((float)i, NTL_WaveFIFO[NTL_Wave_rPtr]);
                                NTL_Wave_rPtr--;
                                if (NTL_Wave_rPtr < 0) { NTL_Wave_rPtr = NTL_WaveFIFO.Length - 1; }
                            }
                            g.DrawLines(new Pen(Color.DeepPink, 3.0f), p1);
                        }
                    }

                    g.Dispose();
                    this.Validate();
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        private string CheckEstimationResult(int sn, NTL_EstimationCircle.NTL_FigureType t)
        {   // <sn> 0:右手 1:左手
            string s = "";
            if (sn == 0) { DetectCountR++; } else { DetectCountL++; }
            if (((sn == 0) && (DetectCountR > 150)) || ((sn == 1) && (DetectCountL > 150))) //Line検出後５秒間は不感状態とする。
            {
                if (sn == 0) { DetectCountR = 150; } else { DetectCountL = 150; }
                if (sn == 0) { s = "右手："; } else { s = "左手："; }
                switch (t)
                {
                    case NTL_EstimationCircle.NTL_FigureType.Circle:
                        s += "円が検出されました"; break;
                    case NTL_EstimationCircle.NTL_FigureType.Arc:
                        s += "円弧が検出されました"; break;
                    case NTL_EstimationCircle.NTL_FigureType.Line:
                        if (sn == 0) { DetectCountR = 0; } else { DetectCountL = 0; }
                        s += "線分が検出されました"; break;
                    case NTL_EstimationCircle.NTL_FigureType.Point:
                        s += "点が検出されました"; break;
                }
            }
            else
            {
                if (sn == 0) { s = "右手：線分が検出されました"; } else { s = "左手：線分が検出されました"; }
            }
            if (sn == 0) { FigR = t; } else { FigL = t; }
            return s;
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void DepthImageReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (KS_DepthFrame = e.OpenDepthImageFrame())
            {
                if (KS_DepthFrame != null)
                {
                    KS_DepthFrame.CopyPixelDataTo(KS_DepthData); //【KINECTのDepth情報を配列DepthDataにコピー】
                    KS_ConvertDepthToColor(); //【深度情報をカラーバーの色を使って表示形式をX8R8G8B8形式に変換】
                    KS_data2 = KS_bmp2.LockBits(KS_rect2, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lockする】
                    Marshal.Copy(KS_DepthImageDataXRGB, 0, KS_data2.Scan0, KS_DepthImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列depthDataをコピーする】
                    KS_bmp2.UnlockBits(KS_data2); //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox4.Image = KS_bmp2;  //【pictureBox1に画像表示】
                }
                else
                {
                    //【DepthFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【ConvertDepthToColor】16ビットのDepth値をカラーバーに対応する色に変更
        //***************************************************************
        private void KS_ConvertDepthToColor()
        {
            float a;      //【Normalモード】800～4000mm ⇒ 0.0796875  3200色
            int i, j, d;  //【Nearモード】  400～3000mm ⇒ 0.0980769　2600色
            int f; //【PlayerLost検出用】
            int playerID; //【下位3bitはplayerID】
            if (KS_KinectSensor.DepthStream.Range == DepthRange.Default) { a = 1.000000f; } else { a = 1.230769f; }
            j = 0;
            f = 0;
            for (i = 0; i < KS_DepthData.Length; i++)
            {
                d = KS_DepthData[i];                  //【距離情報】【上位13bitがDepth値】
                playerID = d & KS_PlayerIndexBitmask; //【下位3bitはplayerID】【未使用】
                if (playerID > 0) { f = 1; }
                d = ((d >> KS_PlayerIndexBitmaskWidth) & 0x1FFF);  //【realDepth】0mm～8191mm
                if (d < KS_Lmin) { d = KS_Lmin; } //【近すぎる】
                if (d > KS_Lmax) { d = KS_Lmax; } //【遠すぎる】
                d = (int)((float)(d - KS_Lmin) * a) * 4;
                KS_DepthImageDataXRGB[j] = ColorBar.NTL_ColorElement[d]; //【青】B
                KS_DepthImageDataXRGB[j + 1] = ColorBar.NTL_ColorElement[d + 1]; //【緑】G
                KS_DepthImageDataXRGB[j + 2] = ColorBar.NTL_ColorElement[d + 2]; //【赤】R
                KS_DepthImageDataXRGB[j + 3] = ColorBar.NTL_ColorElement[d + 3]; //【α】
                j = j + 4;
            }
            if (f == 0) { KS_flagPlayer = false; } else { KS_flagPlayer = true; }
        }

        //***************************************************************
        //【SkeletonFrameの取得イベント】SkeletonFrameは6名分。各々がJointsコレクションを持つ。
        //***************************************************************
        private void KinectAllFramesReady(object sender, AllFramesReadyEventArgs e)
        {
            using (KS_SkeletonFrame = e.OpenSkeletonFrame())
            {
                if (KS_SkeletonFrame != null)
                {
                    KS_SkeletonFrame.CopySkeletonDataTo(KS_SkeletonData); //【KINECTのSkeltonデータを配列SkeletonDataにコピー】
                    KS_TimeStamp = KS_SkeletonFrame.Timestamp; //【現在時刻をミリ秒単位で取得】
                    for (int i = 0; i < KS_SkeletonData.Length; i++)
                    {
                        if (KS_SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                        {
                            label3.Text = "Tracking Player No.=" + i.ToString();
                            NTL_SkeletonData.SetBody(KS_KinectSensor, KS_SkeletonData[i], KS_flagFlip); //【骨格情報の保存】
                            if (KS_SkeletonData[i].Joints[JointType.HandRight].TrackingState == JointTrackingState.Tracked)
                            {
                                Ps = KS_SkeletonData[i].Joints[JointType.HandRight].Position; //【右手SkeletonPointを取得】
                                if (KS_flagFlip) { Ps.X = -(Ps.X); } //【必要なら左右を反転】
                                Pc = KS_KinectSensor.MapSkeletonPointToColor(Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                estR.Estimation((float)Pc.X, (float)Pc.Y); //【右手軌跡を追加して円を推定】
                            }

                            if (KS_SkeletonData[i].Joints[JointType.HandRight].TrackingState == JointTrackingState.Tracked)
                            {
                                NTL_HandySensor.Ps = KS_SkeletonData[i].Joints[JointType.HandRight].Position; //【右手】
                                if (KS_flagFlip) { NTL_HandySensor.Ps.X = -(NTL_HandySensor.Ps.X); }
                                NTL_HandySensor.Pd = KS_KinectSensor.MapSkeletonPointToDepth(NTL_HandySensor.Ps, DepthImageFormat.Resolution640x480Fps30);
                                NTL_HandySensor.Pc = KS_KinectSensor.MapSkeletonPointToColor(NTL_HandySensor.Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                NTL_HandySensor.ColorValue = NTL_SensorColorBar.GetColor(MeasureValue); //【センサ値に応じて色を変更する】
                                NTL_HandySensor.SensorValue = MeasureValue;  //【センサの計測値】未使用
                                NTL_FIFO_Push(NTL_HandySensor);
                            }
                            
                            if (KS_SkeletonData[i].Joints[JointType.HandLeft].TrackingState == JointTrackingState.Tracked)
                            {
                                Ps = KS_SkeletonData[i].Joints[JointType.HandLeft].Position;  //【左手SkeletonPointを取得】
                                if (KS_flagFlip) { Ps.X = -(Ps.X); } //【必要なら左右を反転】
                                Pc = KS_KinectSensor.MapSkeletonPointToColor(Ps, ColorImageFormat.RgbResolution640x480Fps30);
                                estL.Estimation((float)Pc.X, (float)Pc.Y); //【左手軌跡を追加して円を推定】
                            }
                            break;
                        }
                    }
                }
                else
                {
                    //【SkeletonFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }

        //***************************************************************
        //【拡張現実センサ用のFIFO】
        //***************************************************************

        public void NTL_FIFO_Initialize(int n)
        {
            NTL_wPtr = 0;
            NTL_rPtr = 0;
            NTL_nFIFO = n;
            NTL_FIFO_Sensor = new NTL_FIFOData[NTL_nFIFO];
        }

        public void NTL_FIFO_Initialize()
        {
            NTL_wPtr = 0;
            NTL_rPtr = 0;
            NTL_nFIFO = NTL_default_nFIFO;
            NTL_FIFO_Sensor = new NTL_FIFOData[NTL_nFIFO];
        }

        public void NTL_FIFO_Push(NTL_FIFOData Sensor)
        {
            NTL_FIFO_Sensor[NTL_wPtr] = Sensor;
            NTL_wPtr++;
            NTL_FIFOempty = false;
            if (NTL_wPtr == NTL_nFIFO) { NTL_wPtr = 0; NTL_FIFOfull = true; }
        }

        public NTL_FIFOData NTL_FIFO_Read_Sensor()
        {
            NTL_FIFOData a = NTL_FIFO_Sensor[NTL_rPtr];
            NTL_rPtr++;
            if (NTL_rPtr == NTL_nFIFO) { NTL_rPtr = 0; }
            return a;
        }

        //***********************************************************************
        // 時間待ち(ミリ秒単位)
        //***********************************************************************
        [DllImport("kernel32.dll")]
        public static extern void Sleep(Int32 dwMilliseconds);

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板USB-ADCの通信初期化処理】
        //***********************************************************************
        public Boolean InitializeSerialPort(string com)
        {
            //【ADC用シリアルポートの設定】
            this.SerialPort_for_ADC.PortName = COMPORT;
            // トラ技増刊号の付録基板は疑似シリアルポート
            //     this.SerialPort_for_ADC.BaudRate = 115200;
            //     this.SerialPort_for_ADC.DataBits = 8;
            //     this.SerialPort_for_ADC.StopBits = System.IO.Ports.StopBits.One;
            //     this.SerialPort_for_ADC.Parity = System.IO.Ports.Parity.None;
            //     this.SerialPort_for_ADC.ReadTimeout = 50;
            //     this.SerialPort_for_ADC.ReadBufferSize = 256;
            //     this.SerialPort_for_ADC.WriteBufferSize = 256;
            try { this.SerialPort_for_ADC.Open(); }
            catch (Exception) { MessageBox.Show("シリアルポートが開けません"); return false; }
            return true;
        }

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板USB-ADCの通信終了処理】
        //***********************************************************************
        //　ポートを解放する
        public void TerminateSerialPort()
        {
            this.SerialPort_for_ADC.Close();
        }

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板USB-ADCからのAD変換値受信イベント処理】
        //***********************************************************************
        void SerialPort_for_ADC_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            ADCGetValue(); //【シリアルポートからデータを読み込む】
            if (NTL_Ch != NTL_pCh) //【入力チャンネルに変更があった場合】
            {
                ADCInitialize();
                NTL_pCh = NTL_Ch;
            }
            ADCConvert();  //【次のA/D変換データを要求する】
        }

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板】
        //***********************************************************************
        // ■■■■■■■■■■　【A/Dコンバータの初期化処理】　■■■■■■■■■■
        public void ADCInitialize()
        {
            //【接続確認】特に値のチェックなどはしていません。
            this.SerialPort_for_ADC.Write("E\r\n");       //【送信】"E"+CR+LF
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADCモードレジスタ書込み】000A　ADC変換レートを(16.7Hzに)設定する
            // this.SerialPort_for_ADC.Write("C080001\r\n"); //【送信】"C08"+"0001"+CR+LF 500Hz
            // this.SerialPort_for_ADC.Write("C080002\r\n"); //【送信】"C08"+"0002"+CR+LF 250Hz
            // this.SerialPort_for_ADC.Write("C080003\r\n"); //【送信】"C08"+"0003"+CR+LF 125Hz
            // this.SerialPort_for_ADC.Write("C080004\r\n"); //【送信】"C08"+"0004"+CR+LF 62.5Hz
            // this.SerialPort_for_ADC.Write("C080005\r\n"); //【送信】"C08"+"0005"+CR+LF 50Hz
            // this.SerialPort_for_ADC.Write("C080006\r\n"); //【送信】"C08"+"0006"+CR+LF 39.2Hz
            this.SerialPort_for_ADC.Write("C080007\r\n"); //【送信】"C08"+"0007"+CR+LF 33.3Hz
            // this.SerialPort_for_ADC.Write("C080008\r\n"); //【送信】"C08"+"0008"+CR+LF 19.6Hz
            // this.SerialPort_for_ADC.Write("C08000A\r\n"); //【送信】"C08"+"000A"+CR+LF 16.7Hz
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADC設定レジスタ書込み】5080 【コマンド】C105080 + CR + LF
            if (radioButton3.Checked) { this.SerialPort_for_ADC.Write("C105080\r\n"); NTL_Ch = 0; } //【入力】チャンネル0
            if (radioButton4.Checked) { this.SerialPort_for_ADC.Write("C105081\r\n"); NTL_Ch = 1; } //【入力】チャンネル1
            if (radioButton5.Checked) { this.SerialPort_for_ADC.Write("C105082\r\n"); NTL_Ch = 2; } //【入力】チャンネル2
            if (radioButton6.Checked) { this.SerialPort_for_ADC.Write("C105086\r\n"); NTL_Ch = 3; } //【入力】内蔵温度センサ
            if (radioButton7.Checked) { this.SerialPort_for_ADC.Write("C105080\r\n"); NTL_Ch = 4; } //【入力】金属検出センサ
            NTL_pCh = NTL_Ch;
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【IOレジスタ】0A　【コマンド】C280A + CR + LF
            this.SerialPort_for_ADC.Write("C280A\r\n");   //【送信】電流源は未使用
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            //【ADCオフセットレジスタ】800000 【コマンド】C30800000 + CR + LF
            this.SerialPort_for_ADC.Write("C30800000\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PM80\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PW78\r\n"); //【送信】
            Sleep(3000);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.Write("PW00\r\n"); //【送信】
            Sleep(50);                                    //【待機】50m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
            this.SerialPort_for_ADC.DiscardInBuffer();    //【バッファ内データを破棄する】
        }

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板】
        //***********************************************************************
        // ■■■■■■■■■■　【A/Dコンバータの終了処理】　■■■■■■■■■■
        public void ADCTerminate()
        {
            this.SerialPort_for_ADC.Write("C08600A\r\n"); //【送信】パワーダウンモード
            Sleep(5);                                     //【待機】5m秒
            RxMsg = this.SerialPort_for_ADC.ReadLine();   //【受信】
        }

        //***********************************************************************
        //【CQ出版トランジスタ技術増刊号付録基板】
        //***********************************************************************
        // ■■■■■■■■■■　【A/D変換処理】　■■■■■■■■■■
        // ここにA/D変換読み取り処理を挿入する。設定範囲を0≦MeasureValue≦255とする。
        void ADCConvert()
        {
            this.SerialPort_for_ADC.Write("R\r\n");     //【送信】AD変換結果を取得する
        }

        void ADCGetValue()
        {
            const float L1 = 1.17f / 16777216.0f;
            const float L2 = 1.0f / 0.00081f;
            const float L3 = 273.15f; //【温度の単位を[K]から[℃]に変換するための係数】
            ADcount = 0;
            try { RxMsg = this.SerialPort_for_ADC.ReadLine(); } //【受信】
            catch { RxMsg = "000000"; }
            if (RxMsg.Substring(RxMsg.Length - 1, 1) == "\n") { RxMsg = RxMsg.Substring(0, RxMsg.Length - 2); }
            if (RxMsg.Substring(RxMsg.Length - 1, 1) == "\r") { RxMsg = RxMsg.Substring(0, RxMsg.Length - 1); }
            if (RxMsg.Substring(0, 1) == "\0") { RxMsg = RxMsg.Substring(1, RxMsg.Length - 1); }
            ADcount = Convert.ToInt32(RxMsg, 16); //★Convertクラスを使って16進数を符号付き整数として変換する
            SensorValue = L1 * (float)ADcount; //【まず測定電圧に変換する】
            switch (NTL_Ch)
            {
                case 0:
                    SensorValue *= 3.0f; //【超音波センサの場合】
                    sSensorValue = SensorValue.ToString() + "[V]";
                    break;
                case 1:
                    SensorValue = 1.0f - SensorValue; //【CdS照度センサ】
                    sSensorValue = SensorValue.ToString() + "[V]";
                    break;
                case 2:
                    SensorValue = (SensorValue - 0.40f) * 5.5f;
                    sSensorValue = SensorValue.ToString() + "[V]";
                    break;
                case 3:
                    float Temperature;
                    Temperature = SensorValue * L2 - L3;
                    sSensorValue = Temperature.ToString() + "[℃]";
                    SensorValue = (Temperature - 10.0f) * 0.025f; //10℃から50℃までを0.0から1.0に変換する
                    break;
                case 4:
                    SensorValue = (SensorValue - 0.250f) * 2.0f;
                    sSensorValue = SensorValue.ToString() + "[V]";
                    break;
            }
            MeasureValue = (int)(SensorValue * 255.0f);
            if (MeasureValue < 0) { MeasureValue = 0; }
            if (MeasureValue > 255) { MeasureValue = 255; }
        }
    }
}
